@@ -2,17 +2,22 @@ package Net::DNS::Resolver::Unbound;
 
 use strict;
 use warnings;
+use integer;
 use Net::DNS;
-use base qw(Net::DNS::Resolver DynaLoader);
+
+use constant OS_SPEC => defined eval "require Net::DNS::Resolver::$^O";	## no critic
+use constant OS_CONF => join '::', 'Net::DNS::Resolver', OS_SPEC ? $^O : 'UNIX';
+use base qw(Net::DNS::Resolver::Base DynaLoader), OS_CONF;
 
 our $VERSION;
 
 BEGIN {
-	$VERSION = '1.19_03';
+	$VERSION = '1.24_01';
 	eval { __PACKAGE__->bootstrap($VERSION) };
 }
 
-use constant UB_SEND => Net::DNS::Resolver::Unbound::Context->can('ub_send');
+use constant UB_CONTEXT => 'Net::DNS::Resolver::Unbound::Context';
+use constant UB_SEND => UB_CONTEXT->can('ub_send');
 
 
 =head1 NAME
@@ -41,8 +46,8 @@ As of this writing, the implementation has some significant limitations:
 
 =item *
 
-Selection of transport protocol and associated parameters is entirely
-at the discretion of Unbound.
+Selection of transport protocol and associated parameters is almost
+entirely at the discretion of Unbound.
 
 =item *
 
@@ -55,6 +60,11 @@ It is not possible to send a pre-constructed packet to a nameserver.
 A best-effort attempt is made instead using (qname,qtype,qclass)
 extracted from the presented packet.
 
+=item *
+
+Result packet is synthesised in libunbound and not the "real thing".
+In particular, the returned queryID is always zero.
+
 =back
 
 
@@ -68,22 +78,26 @@ extracted from the presented packet.
 	dnsrch,	    => 1,
 	domain	    => 'domain',
 	ndots	    => 1,
-	searchlist  => ['domain' ... ],
+	option	    => [ 'tls-cert-bundle', '/etc/ssl/cert.pem' ],
 	nameservers => [ ... ],
-	option => ['logfile', 'mylog.txt'] );
+	searchlist  => ['domain' ... ],
+	);
 
 Returns a new Net::DNS::Resolver::Unbound resolver object.
 
 =cut
 
 sub new {
-	my $class = shift;
-	my $self  = $class->SUPER::new();
-	$self->{ub_ctx} = Net::DNS::Resolver::Unbound::Context->new();
-	while ( my $attr = shift ) {
-		my $value = shift;
+	my ( $class, @args ) = @_;
+	my $self = $class->SUPER::new();
+	$self->nameservers( $self->SUPER::nameservers );
+	$self->_finalise_config;				# default configuration
+	$self->{update} = {} if @args;				# force context rebuild
+	while ( my $attr = shift @args ) {
+		my $value = shift @args;
 		$self->$attr( ref($value) ? @$value : $value );
 	}
+	$self->_finalise_config;
 	return $self;
 }
 
@@ -103,7 +117,7 @@ sub new {
 	add_ta_file => '/var/lib/unbound/root.key'
 	);
 
-    my @nameservers = $stub_resolver->nameservers;
+    $stub_resolver->nameservers( '127.0.0.53', ... );
 
 By default, DNS queries are sent to the IP addresses listed in
 /etc/resolv.conf or similar platform-specific sources.
@@ -111,11 +125,19 @@ By default, DNS queries are sent to the IP addresses listed in
 =cut
 
 sub nameservers {
-	my $self = shift;
-	local $self->{debug};		## "no nameservers" ok in this context
-	return $self->SUPER::nameservers(@_);
+	my ( $self, @nameservers ) = @_;
+	if ( defined wantarray ) {
+		$self->_finalise_config;
+		my %config = %{$self->{config}};
+		return @{$config{set_fwd}};
+	}
+	$self->set_fwd() unless @nameservers;
+	$self->set_fwd($_) foreach @nameservers;
+	return;
 }
 
+sub nameserver { &nameservers; }
+
 
 =head2 search, query, send, bgsend, bgbusy, bgread
 
@@ -124,24 +146,33 @@ See L<Net::DNS::Resolver>.
 =cut
 
 sub send {
-	my $self = shift;
+	my ( $self, @argument ) = @_;
 	$self->_finalise_config;
 	$self->_reset_errorstring;
 
-	my $query = $self->_make_query_packet(@_);
-	my ($q) = $query->question;
-	my $reply = $self->{ub_ctx}->ub_resolve( $q->name, $q->{qtype}, $q->{qclass} );
-	return $self->_decode_result( $reply, $query->header->id );
+	my ($packet) = @argument;
+	my $query = $self->_make_query_packet(@argument);
+	my $result;
+	if ( UB_SEND && ref($packet) ) {
+		$result = $self->{ub_ctx}->ub_send( $query->encode );
+	} else {
+		my ($q) = $query->question;
+		$result = $self->{ub_ctx}->ub_resolve( $q->name, $q->{qtype}, $q->{qclass} );
+	}
+
+	return $self->_decode_result($result);
 }
 
 sub bgsend {
-	my $self = shift;
+	my ( $self, @argument ) = @_;
 	$self->_finalise_config;
 	$self->_reset_errorstring;
 
-	my $query = $self->_make_query_packet(@_);
-	my ($q) = $query->question;
-	return $self->{ub_ctx}->ub_resolve_async( $q->name, $q->{qtype}, $q->{qclass}, $query->header->id );
+	my $query = $self->_make_query_packet(@argument);
+	my $image = $query->encode;
+	my $ident = $query->header->id;
+	my ($q)	  = $query->question;
+	return $self->{ub_ctx}->ub_resolve_async( $q->name, $q->{qtype}, $q->{qclass}, $ident );
 }
 
 sub bgbusy {
@@ -160,24 +191,33 @@ sub bgread {
 	$self->{ub_ctx}->ub_wait if &bgbusy;
 
 	$self->errorstring( $handle->err );
-	return $self->_decode_result( $handle->result, $handle->query_id );
+	my $reply = $self->_decode_result( $handle->result ) || return;
+	$reply->header->id( $handle->query_id );
+	return $reply;
 }
 
 
 =head2 option
 
-    $filename = $resolver->option( 'logfile' );
-    $resolver->option( 'logfile', $filename );
+    $resolver->option( 'tls-cert-bundle', '/etc/ssl/cert.pem' );
 
-Get or set Unbound resolver (name,value) context options.
+Set Unbound resolver (name,value) context option.
 
 =cut
 
 sub option {
 	my ( $self, $name, @value ) = @_;
-	return $self->{ub_ctx}->set_option( "$name:", @value ) if @value;
-	my $value = $self->{ub_ctx}->get_option($name);
-	return wantarray ? split /\r*\n/, $value : $value;
+	unless (@value) {
+		$self->_finalise_config;
+		my @return = eval {
+			my $optref = $self->{config}->{set_option};
+			my %option = @$optref;
+			my @list = @{$option{"${name}:"}};
+		};
+		return wantarray ? @return : shift @return;
+	}
+
+	return $self->_option( $name, @value );
 }
 
 
@@ -192,7 +232,7 @@ Unbound configuration options.
 
 sub config {
 	my ( $self, $filename ) = @_;
-	return $self->{ub_ctx}->config($filename);
+	return $self->_config( 'config', $filename );
 }
 
 
@@ -209,8 +249,8 @@ as backup servers.
 =cut
 
 sub set_fwd {
-	my ( $self, $fwd ) = @_;
-	return $self->{ub_ctx}->set_fwd($fwd);
+	my ( $self, @fwd ) = @_;
+	return $self->_config( 'set_fwd', [@fwd] );
 }
 
 
@@ -223,9 +263,11 @@ Use DNS over TLS for queries to nameservers specified using set_fwd().
 
 =cut
 
+
 sub set_tls {
-	my ( $self, $do_tls ) = @_;
-	return $self->{ub_ctx}->set_tls($do_tls);
+	my ( $self, $tls ) = @_;
+	use constant SET_TLS => UB_CONTEXT->can('set_tls');
+	return $self->_config( 'set_tls', $tls ) if SET_TLS;
 }
 
 
@@ -241,7 +283,8 @@ and the 'DHCP DNS' IP address, use set_fwd().
 
 sub set_stub {
 	my ( $self, $zone, $address, $prime ) = @_;
-	return $self->{ub_ctx}->set_stub( $zone, $address, $prime );
+	use constant SET_STUB => UB_CONTEXT->can('set_stub');
+	return $self->_config( 'set_stub', [[$zone, $address, $prime]] ) if SET_STUB;
 }
 
 
@@ -259,7 +302,7 @@ or other platform-specific sources.
 
 sub resolv_conf {
 	my ( $self, $filename ) = @_;
-	return $self->{ub_ctx}->resolv_conf($filename);
+	return $self->_config( 'resolv_conf', $filename );
 }
 
 
@@ -274,7 +317,7 @@ These addresses are not flagged as DNSSEC secure when queried.
 
 sub hosts {
 	my ( $self, $filename ) = @_;
-	return $self->{ub_ctx}->hosts($filename);
+	return $self->_config( 'hosts', $filename );
 }
 
 
@@ -288,8 +331,9 @@ in RFC1035 zonefile format.
 =cut
 
 sub add_ta {
-	my $self = shift;
-	return $self->{ub_ctx}->add_ta( Net::DNS::RR->new(@_)->plain );
+	my ( $self, @argument ) = @_;
+	my $ta = Net::DNS::RR->new(@argument)->plain;
+	return $self->_config( 'add_ta', $ta );
 }
 
 
@@ -304,7 +348,7 @@ Pass the name of a file containing DS and DNSKEY records
 
 sub add_ta_file {
 	my ( $self, $filename ) = @_;
-	return $self->{ub_ctx}->add_ta_file($filename);
+	return $self->_config( 'add_ta_file', $filename );
 }
 
 
@@ -320,7 +364,8 @@ trust anchor is changed.
 
 sub add_ta_autr {
 	my ( $self, $filename ) = @_;
-	return $self->{ub_ctx}->add_ta_autr($filename);
+	use constant ADD_TA_AUTR => UB_CONTEXT->can('add_ta_autr');
+	return $self->_config( 'add_ta_autr', $filename ) if ADD_TA_AUTR;
 }
 
 
@@ -334,7 +379,7 @@ Pass the name of a BIND-style config file containing trusted-keys{}.
 
 sub trusted_keys {
 	my ( $self, $filename ) = @_;
-	return $self->{ub_ctx}->trusted_keys($filename);
+	return $self->_config( 'trusted_keys', $filename );
 }
 
 
@@ -348,8 +393,8 @@ Pass a null argument to disable. Default is stderr.
 =cut
 
 sub debug_out {
-	my ( $self, $out ) = @_;
-	return $self->{ub_ctx}->debug_out($out);
+	my ( $self, $stream ) = @_;
+	return $self->_config( 'debug_out', $stream );
 }
 
 
@@ -365,7 +410,7 @@ Set verbosity of the debug output directed to stderr.  Level 0 is off,
 sub debug_level {
 	my ( $self, $verbosity ) = @_;
 	$self->debug($verbosity);
-	return $self->{ub_ctx}->debug_level($verbosity);
+	return $self->_config( 'debug_level', $verbosity );
 }
 
 
@@ -381,8 +426,8 @@ If false (by default), a process is forked to perform the work.
 =cut
 
 sub async_thread {
-	my ( $self, $dothread ) = @_;
-	return $self->{ub_ctx}->async($dothread);
+	my ( $self, $threaded ) = @_;
+	return $self->_config( 'async', $threaded );
 }
 
 
@@ -395,32 +440,46 @@ Prints the resolver state on the standard output.
 
 =cut
 
-sub print {
-	return print shift->string;
-}
-
 sub string {
 	my $self = shift;
-	$self = $self->_defaults unless ref($self);
+	$self = $self->new() unless ref($self);
 
-	my @nslist   = $self->nameservers();
-	my ($force)  = ( grep( { $self->{$_} } qw(force_v6 force_v4) ),	  'force_v4' );
-	my ($prefer) = ( grep( { $self->{$_} } qw(prefer_v6 prefer_v4) ), 'prefer_v4' );
-	return <<END;
+	my $image = <<END;
 ;; RESOLVER state:
-;; nameservers	= @nslist
-;; searchlist	= @{$self->{searchlist}}
-;; defnames	= $self->{defnames}	dnsrch		= $self->{dnsrch}
-;; ${prefer}	= $self->{$prefer}	${force}	= $self->{$force}
-;; debug	= $self->{debug}	ndots		= $self->{ndots}
+;; debug	$self->{debug}	ndots	$self->{ndots}
+;; defnames	$self->{defnames}	dnsrch	$self->{dnsrch}
+;; searchlist	@{$self->{searchlist}}
 END
+	$self->_finalise_config;
+	my %config = %{$self->{config}};
+	my $optref = $config{set_option};			# expand option list
+	my %option = @{$optref || []};
+	my @option;
+	foreach my $opt ( sort keys %option ) {
+		push @option, [$opt, $_] for @{$option{$opt}};
+	}
+
+	my $format = ";; %s\t%s\n";
+	foreach my $name ( sort keys %config ) {
+		local $config{set_option} = \@option;
+		my $value = $config{$name};
+		if ( ref $value ) {
+			foreach my $arg (@$value) {
+				my @arg = map { ref($_) ? @$_ : $_ } $arg;
+				$image .= sprintf( $format, $name, join ' ', @arg );
+			}
+		} else {
+			$image .= sprintf( $format, $name, $value );
+		}
+	}
+	return $image;
 }
 
 
 ########################################
 
 sub _decode_result {
-	my ( $self, $result, $query_id ) = @_;
+	my ( $self, $result ) = @_;
 
 	return unless $result;
 
@@ -428,45 +487,105 @@ sub _decode_result {
 	$self->errorstring( $result->why_bogus ) if $result->bogus;
 
 	my $buffer = $result->answer_packet || return;
-	my $packet = Net::DNS::Packet->decode(\$buffer);
+	my $packet = Net::DNS::Packet->decode( \$buffer );
 	$self->errorstring($@);
-	$packet->header->id($query_id) if !UB_SEND || $query_id;
 	$packet->print if $self->debug;
 
 	return $packet;
 }
 
 
-sub _finalise_config {
-	my $self = shift;
-	return if $self->{ub_frozen}++;
-
-	my %IP_conf = (
-		force_v4  => ['do-ip6'	   => 'no'],
-		force_v6  => ['do-ip4'	   => 'no'],
-		prefer_v4 => ['prefer-ip4' => 'yes'],
-		prefer_v6 => ['prefer-ip6' => 'yes'] );
-
-	for ( grep { $self->{$_} } qw(prefer_v4 prefer_v6 force_v4 force_v6) ) {
-		my $argref = $IP_conf{$_};
-		eval { $self->option(@$argref) };		# unimplemented in old versions
-	}
-
-	my $count = 3;
-	foreach ( grep { $count-- > 0 } $self->nameservers ) {
-		$self->set_fwd($_);
+sub _config {
+	my ( $self, $name, $arg ) = @_;
+	$self->{ub_ctx} = Net::DNS::Resolver::Unbound::Context->new() unless $self->{update};
+	my $update = $self->{update} ||= {};			# collect context changes
+	if ( ref $arg ) {
+		my @arg = map { ref($_) ? @$_ : $_ } @$arg;
+		$self->{ub_ctx}->$name(@arg) if @arg;		# error check only
+		my $list = $update->{$name} ||= [];
+		push @$list, @$arg if @arg;			# append non-empty
+		$update->{$name} = [] unless @arg;		# explicit empty list
+	} elsif ( defined $arg ) {
+		$self->{ub_ctx}->$name($arg);			# error check only
+		$update->{$name} = $arg;			# [re]define scalar
 	}
 	return;
 }
 
 
+sub _option {
+	my ( $self, $name, $arg ) = @_;
+	$self->{ub_ctx} = Net::DNS::Resolver::Unbound::Context->new() unless $self->{update};
+	my $setopt = $self->{config}->{set_option};
+	my $optref = $self->{update}->{set_option} || $setopt;
+	my %option = @{$optref || []};
+
+	my $opt = "${name}:";
+	if ($arg) {
+		my $list = $option{$opt} ||= [];
+		$self->{ub_ctx}->set_option( $opt, $arg );
+		push @$list, $arg;
+		$self->{update}->{set_option} = [%option];
+	}
+
+	return;
+}
+
+
+my %IP_conf = (
+	force_v4  => ['do-ip6'	   => 'no'],
+	force_v6  => ['do-ip4'	   => 'no'],
+	prefer_v4 => ['prefer-ip4' => 'yes'],
+	prefer_v6 => ['prefer-ip6' => 'yes'] );
+
+sub _finalise_config {
+	my $self = shift;
+
+	return unless $self->{update};
+
+	foreach my $key ( keys %IP_conf ) {
+		my ( $name, $arg ) = @{$IP_conf{$key}};
+		my @arg = $self->$key ? $arg : ();
+		eval { $self->_option( $name, @arg ); 1 }	# if implemented
+	}
+
+	my $config = $self->{config};
+	my $update = delete $self->{update};
+	my %config = ( %$config, %$update );
+
+	my $optref = $config{set_option};			# expand option list
+	my %option = @{$optref || []};
+	my @option;
+	foreach my $opt ( keys %option ) {
+		push @option, [$opt, $_] for @{$option{$opt}};
+	}
+
+	my $ctx = $self->{ub_ctx} = Net::DNS::Resolver::Unbound::Context->new();
+
+	foreach my $name ( keys %config ) {			# rebuild unbound context
+		local $config{set_option} = \@option;
+		my $value = $config{$name};
+		if ( ref $value ) {
+			foreach my $arg (@$value) {
+				my @arg = map { ref($_) ? @$_ : $_ } $arg;
+				$ctx->$name(@arg);
+			}
+		} else {
+			$ctx->$name($value);
+		}
+	}
+
+	$self->{config} = \%config;
+	return;
+}
+
 1;
 __END__
 
 
 =head1 COPYRIGHT
 
-Copyright (c)2022 Dick Franks
+Copyright (c)2022,2023 Dick Franks
 
 All Rights Reserved
 
